import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Graphical representation of a MIDI note.
 * 
 * @author  Andersson, Andreen, Lanneskog, Pehrson
 * @version 1
 */
 
public class MooNoteElement extends JPanel {

	private MooTrackView mtv;
	private MooNote note;
	private JPopupMenu popup;
	private JMenuItem popupRemove, popupProp, popupTranspOctUp, popupTranspOctDown;
	private Rectangle pitchRect, veloRect;
	private String notePitch, noteVelocity;
	private boolean selected = false;
	public Color textColor;
	public static final Color bgColor = new Color(160, 218, 255);
	public static final Color invBgColor = new Color(96, 38, 0);

	/** 
	 * Creates a new note element.
	 * @param parent	The MooTrackView that this element will be painted on.
	 * @param mn		the note that will be graphically represented
	 */
	public MooNoteElement (MooTrackView parent, MooNote mn) {
		mtv = parent;
		note = mn;
		calculateString();
		addMouseListener(new MAdapter());
		setBorder(BorderFactory.createLineBorder(Color.black));
		setBackground(bgColor);
		textColor = Color.black;

		// Defines coordinates.
		pitchRect = new Rectangle(0, 0, 15, 10);
		veloRect = new Rectangle(20, 0, 40, 10);

		// Creates pop-up menu.
		popup = new JPopupMenu();
		PopupListener pList = new PopupListener();
		popupProp = new JMenuItem("Preferences...");
		popupProp.addActionListener(pList);
		popup.add(popupProp);
		popupRemove = new JMenuItem("Remove");
		popupRemove.addActionListener(pList);
		popup.add(popupRemove);
		popupTranspOctUp = new JMenuItem("Transpose one octave up");
		popupTranspOctUp.addActionListener(pList);
		popup.add(popupTranspOctUp);
		popupTranspOctDown = new JMenuItem("Transpose one octave down");
		popupTranspOctDown.addActionListener(pList);
		popup.add(popupTranspOctDown);
	}

	/** 
	 * Returns the note of this element.
	 * @return the note
	 */
	public MooNote getNote() {
		return note;
	}

	/** 
	 * Selects the current NoteElement.
	 */
	public void select() {
		selected = true;
		mtv.selectNote(this);
		setBackground(invBgColor);
		textColor = Color.white;
		repaint();
	}

	/** 
	 * Deselects the current NoteElement.
	 */
	public void deselect() {
		selected = false;
		// mtv.deselectNote(this);
		setBackground(bgColor);
		textColor = Color.black;
		repaint();
	}

	/** 
	 * Transposes the current note element the given number of halftones.
	 * @param halftones	the number of halftones to transpose - positive for up, negative for down
	 */
	public void transpose(int halftones) {
		note.transpose(halftones);
		update();
	}

	/**
	 * Draws the string that shows the note's properties.
	 * @param g	The Graphics object used to draw the strings.
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		if (!(g instanceof Graphics2D)) return;
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(textColor);
		g2.setFont(new Font("Helvetica", Font.PLAIN, 8));
	/*
		switch(columns) {
			case 0:
			case 1:
			...
		}
	*/
		
		g2.drawString(notePitch, 1, 8);
		g2.drawString("" + noteVelocity, 21, 8);
	}

	/**
	 * Calculate what the string that shows the note properties should look like.
	 */
	protected void calculateString(){

		noteVelocity = ""; 
		notePitch = "";
		if(note == null) return;

		int pitch = note.getPitch();
		switch (pitch % 12) {
			case  0: notePitch = "C";  break;
			case  1: notePitch = "C#"; break;
			case  2: notePitch = "D";  break;
			case  3: notePitch = "D#"; break;
			case  4: notePitch = "E";  break;
			case  5: notePitch = "F";  break;
			case  6: notePitch = "F#"; break;
			case  7: notePitch = "G";  break;
			case  8: notePitch = "G#"; break;
			case  9: notePitch = "A";  break;
			case 10: notePitch = "A#"; break;
			case 11: notePitch = "B";  break;
		}
		notePitch += pitch / 12 - 1;
		noteVelocity = ""+note.getVelocity();
	}


	/**
	 * Asks the MooTrackView that the note element is painted on to remove this element and the note.
	 */
	protected void remove(){
		mtv.removeNote(this);
	}

	/**
	 * Updates the graphical content of the element and repaints it.
	 */
	public void update() {
		calculateString();
		repaint();
	}

	/**
	 * Layout this changed elemnt.
	 */
	protected void newLayout(){
		mtv.layoutElement(this,true);
	}

	/**
	 * Listener that checks the mouse actions on this element.
	 */
	class MAdapter extends MouseAdapter {
	
		/**
		 * If left mouse button is clicked, selects the note and plays it.
		 */
		public void mouseClicked(MouseEvent e) {
			if (SwingUtilities.isLeftMouseButton(e) && !e.isControlDown()) {
				select();
				// Play the note
			}
		}

		/**
		 * Checks if the mouse is pressed.
		 * Increases the pitch or velocity if the right mouse button is pressed while holding CTRL.
		 * Decreases the pitch or velocity if the left mouse button is pressed while holding CTRL.
		 * @param e the event recieved.
		 */
		public void mousePressed(MouseEvent e) {
			if (e.isControlDown()) {
				if (pitchRect.contains(e.getPoint())) {
					if (SwingUtilities.isRightMouseButton(e)) {
						note.transpose(1);
					} else if (SwingUtilities.isLeftMouseButton(e)) {
						note.transpose(-1);
					}
					Moosique.setEdited();
					calculateString();
					repaint();
				} else if (veloRect.contains(e.getPoint())) {
					if (SwingUtilities.isRightMouseButton(e)) {
						note.setVelocity(note.getVelocity() + 1);
					} else if (SwingUtilities.isLeftMouseButton(e)) {
						note.setVelocity(note.getVelocity() - 1);
					}
					Moosique.setEdited();
					calculateString();
					repaint();
				}
			} else maybeShowPopup(e);
		}

		public void mouseReleased(MouseEvent e) {
			maybeShowPopup(e);
		}

		/**
		 * Shows the menu if an OS-specific popup-trigger was activated.
		 */
		private void maybeShowPopup(MouseEvent e) {
			if (e.isPopupTrigger() && !e.isControlDown()) {
				if (!selected || mtv.isTheOnlySelected((MooNoteElement)e.getComponent())) popup.show(e.getComponent(), e.getX(), e.getY());
				else mtv.showSelectionPopup(e.getComponent(), e.getX(), e.getY());
			}
		}
	}

	/**
	 * Listens on the actions made to the popupmenu.
	 */
	class PopupListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if (source == popupProp) {
				new MooDialog(note);
				newLayout();
			} else if (source == popupRemove) {
				remove();
			} else if (source == popupTranspOctUp) {
				transpose(12);
			} else if (source == popupTranspOctDown) {
				transpose(-12);
			}
		}
	}
}
