import javax.sound.midi.*;
import java.io.*;
import javax.swing.*;
import java.util.*;

/**
 * Moosique - The MIDI Tracker
 * 
 * Main class that handles initiation, IO and sound.
 * 
 * @author  Einar Pehrson
 */
 
public class Moosique {

	private static MooGUI gui;
	private static Sequence seq;
	private static Sequencer sequencer;
	private static Synthesizer synthesizer;
	private static Receiver receiver;
	private static MidiChannel[] channels;
	private static MidiChannel activeChannel;
	private static MidiEvent[] timeSignatures, tempoChanges;
	private static ArrayList emptyTracks;
	private static Map trackMute = new HashMap();
	private static Map trackSolo = new HashMap();


	private static String filename;
	private static long editPosition;
	private static boolean makeGUI = true, initSound = true, isEdited = false, drawEmptyTracks = false;
	private static Thread player;
	public static final int DEFAULT_RESOLUTION = 96, DEFAULT_TRACKS = 4;

	/** 
	 * Starts the application.
	 * 
	 * Parses command-line arguments, acquires MIDI devices and connects them,
	 * loads a sequence and creates the GUI.
	 */
	public static void main (String[] args) {
		System.out.println("\nMoosique version 1.0\n");

		// Parses command-line arguments.
		String fileArg = null;
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-n")) makeGUI = false;
			else if (args[i].equals("-m")) initSound = false;
			else if (fileArg == null) fileArg = args[i];
		}

		// Acquires MIDI devices and connects them.
		System.out.print("Initializing MIDI devices.");
		try {
			sequencer = MidiSystem.getSequencer();
			System.out.print(".");
			sequencer.open();
			synthesizer = MidiSystem.getSynthesizer();
			System.out.print(".");
			synthesizer.open();
			receiver = synthesizer.getReceiver();
			sequencer.getTransmitter().setReceiver(receiver);
			channels = synthesizer.getChannels();
			setActiveChannel(0);
		} catch (MidiUnavailableException e) {
			System.out.println("Failed, quitting.");
			System.exit(1);
		}
		System.out.println("Done");

		//If a filename is given as the command-line argument, attempts to load a sequence from the file.
		if (fileArg != null) {
			System.out.print("Loading MIDI sequence from " + fileArg + "...");
			if (!load(fileArg)) {
				System.out.println("Failed, creating new sequence");
				clearSequence();
			} else {
				System.out.println("Done");
			}
		} else {
			// Otherwise creates a new empty one.
			clearSequence();
		}

		// Builds GUI, unless n-flag is set.
		if (makeGUI) {
			System.out.print("Building GUI.");
			try {
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			} catch (Exception e) {}
			gui = new MooGUI(seq);
			System.out.println("Done");
		} else {
			System.out.print("Playing...");
			play();
			while (sequencer.isRunning()) {}
			System.out.println("Done");
			quit();
		}
	}








	/*				***
	 **	  ACCESSOR METHODS	 **
	 ***				  */








	/** 
	 * Returns the currently active MidiChannel.
	 * @return the active MidiChannel
	 */
	public static MidiChannel getActiveChannel() {
		return activeChannel;
	}

	/** 
	 * Returns the MidiChannels of the selected synthesizer.
	 * @return the available MidiChannels
	 */
	public static MidiChannel getChannel(int i) {
		return channels[i];
	}

	/** 
	 * Returns the MidiChannels of the selected synthesizer.
	 * @return the available MidiChannels
	 */
	public static MidiChannel[] getChannels() {
		return channels;
	}

	/** 
	 * Returns the current editing position of the sequencer.
	 * @return the tick position
	 */
	public static long getEditPosition() {
		return editPosition;
	}

	/** 
	 * Returns the GUI.
	 * @return the GUI
	 */
	public static MooGUI getGUI() {
		return gui;
	}

	/** 
	 * Calculates the position (measures, beats, ticks) in the current sequence for the given tick position.
	 * @return an array of integers where index 0 is measures, 1 is beats and 2 is ticks.
	 */
	public static int[] getPositionForTick(long ticks) {
		/*
		int measures, beats, ticks;
		for (int i = 0; i < timeSignatures.length; i++) {
			long tick = timeSignatures[i].getTick();
			// Split the ticks in the interval into measures, beats and ticks.
		}
		*/
		int[] pos = {1, 1, 1};
		return pos;
	}

	/** 
	 * Returns the receiver of the current sequencer.
	 * @return the receiver
	 */
	public static Receiver getReceiver() {
		return receiver;
	}

	/** 
	 * Returns the current sequence.
	 * @return the current sequence
	 */
	public static Sequence getSequence() {
		return seq;
	}

	/** 
	 * Returns the current sequencer.
	 * @return the current sequencer
	 */
	public static Sequencer getSequencer() {
		return sequencer;
	}

	/** 
	 * Returns the tempo of the current sequence.
	 * @return the tick position
	 */
	public static int getTempo() {
		return 120;
		// if (tempoMsg == null) return 0;
	}

	/** 
	 * Calculates the tick position in the current sequence for the given position (measures, beats, ticks).
	 * @return the tick position.
	 */
	public static long getTicksForPosition(int measures, int beats, int ticks) {
		long tickPos = 0;
		/*
		for (int i = 0; i < timeSignatures.length; i++) {
			long tick = timeSignatures[i].getTick();
			// Add the measures, beats and ticks in the interval.
		}
		*/
		return tickPos;
	}

	/** 
	 * Returns the tempo of the current sequence.
	 * @return the tick position
	 */
	public static int[] getTimeSig() {
		int[] ts = {4, 4};
		return ts;
		// if (timeSigMsg == null) return 0;
	}

	/** 
	 * Returns true if the current sequence has been edited.
	 * @return the tick position
	 */
	public static boolean isEdited() {
		return isEdited;
	}

	/** 
	 * Returns whether the given track should be drawn
	 * @return true if the given track should be drawn
	 */
	public static boolean shouldBeDrawn(Track track) {
		if (drawEmptyTracks) return true;
		else return (!emptyTracks.contains(track));
	}








	/*				***
	 **	  MUTATOR METHODS	 **
	 ***				  */








	/** 
	 * Fast forwards the current sequence the given number of measures.
	 * @param measures	the number of measures to fast forward
	 */
	public static void forward(long ticks) {
		editPosition += ticks;
	}

	/** 
	 * Rewinds the current sequence the given number of measures.
	 * @param measures	the number of measures to rewind
	 */
	public static void rewind(long ticks) {
		editPosition -= ticks;
	}

	/** 
	 * Sets the currently active MidiChannel.
	 * @param channel	the number of the MidiChannel to activate
	 */
	public static void setActiveChannel(int channel) {
		activeChannel = channels[channel];
	}

	/** 
	 * Sets whether empty tracks should be drawn
	 * @param state		true if empty tracks should be drawn
	 */
	public static void setDrawEmptyTracks(boolean state) {
		drawEmptyTracks = state;
	}

	/** 
	 * Sets the current sequence as edited, which implies prompts when loading a new sequence.
	 */
	public static void setEdited() {
		isEdited = true;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setEditPosition(long ticks) {
		editPosition = ticks;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setTempo(int bpm) {
		// tempoMsg
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setTimeSig(int bpm) {
		// timeSigMsg
	}
	
	public static void setTrackSolo(Track track, boolean on){
		trackSolo.put(track, new Boolean(on));	
	}

	public static void setTrackMute(Track track, boolean on){
		trackMute.put(track, new Boolean(on));	
	}








	/*				***
	 **	  PLAYBACK METHODS	 **
	 ***				  */








	/** 
	 * Starts playback of the current sequence.
	 */
	public static void play() {
		sequencer.setTickPosition(editPosition);
		resume();
	}

	/** 
	 * Resumes playback of the current sequence.
	 */
	public static void resume() {
		gui.update(0);
		try {
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		Track[] tracks = seq.getTracks();

		sequencer.start();

		for (int i = 0; i < tracks.length; i++) {

			Object ob = trackSolo.get(tracks[i]);
			if(ob instanceof Boolean){
				sequencer.setTrackSolo(i,((Boolean)ob).booleanValue());
			}

			ob = trackMute.get(tracks[i]);
			if(ob instanceof Boolean){
				sequencer.setTrackMute(i,((Boolean)ob).booleanValue());
			}
		}


		// Disables input to volatile components
		// gui.disable();

		// Creates the visualisation thread and starts it.
		player = new Thread () {
			public void run() {
				while(sequencer.isRunning()) {
					// Updates the GUI with the current tick position.
					gui.update(sequencer.getTickPosition());

					// Puts the thread to sleep for as long as it takes
					// the sequencer to reach the next sixteenth.
					try {
						//sleep((long)((15000 / getTempo()) * (tickDiff / ticksPerSixteenth)));
						sleep (10);
					} catch (InterruptedException e) {
						Moosique.stop();
					}
				}
				Moosique.stop();
			}
		};
		player.start();
	}

	/** 
	 * Pauses playback of the current sequence.
	 */
	public static void pause() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		if (player != null) player.interrupt();
	}

	/** 
	 * Stops playback of the current sequence.
	 */
	public static void stop() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		sequencer.setTickPosition(editPosition);
		if (player != null) player.interrupt();
		gui.update((long)0);
	}








	/*				***
	 **	  SYSTEM & IO METHODS	 **
	 ***				  */








	/** 
	 * Replaces the current sequence with a new one, holding three empty tracks.
	 */
	public static void clearSequence() {
		// Creates a new sequence and sends it to the sequencer.
		try {
			seq = new Sequence(Sequence.PPQ, DEFAULT_RESOLUTION, DEFAULT_TRACKS);
			sequencer.setSequence(seq);
			filename = null;
			emptyTracks = new ArrayList();
			trackSolo = new HashMap();
			trackMute = new HashMap();
		} catch (InvalidMidiDataException e) {}
		// Sends sequence to GUI.
		if (gui != null) gui.setSequence(seq);
	}

	/** 
	 * Loads a MIDI sequence from the given file.
	 * @param filename	the filename to use
	 */
	public static boolean load(String file) {
		// Loads sequence from file
		filename = file;
		try {
			seq = MidiSystem.getSequence(new File(filename));
		} catch (InvalidMidiDataException e) {
			return false;
		} catch (IOException e) {
			return false;
		}
		isEdited = false;

		Track[] tracks = seq.getTracks();
		emptyTracks = new ArrayList();
		trackMute = new HashMap();
		trackSolo = new HashMap();

		// Searches track 0 for changes in tempo and time signature.
		MidiEvent event;
		MetaMessage metaMsg;
		ArrayList ts = new ArrayList(), tc = new ArrayList();
		for (int i = 0; i < tracks[0].size(); i++) {
			event = tracks[0].get(i);
			if (event.getMessage().getStatus() == MetaMessage.META) {
				metaMsg = (MetaMessage)event.getMessage();
				switch(metaMsg.getType()) {
					case 81: tc.add(event); break;
					case 88: ts.add(event);
				}
			}
		}
		// timeSignatures = ts.toArray(timeSignatures);
		// tempoChanges = tc.toArray(tempoChanges);

		// Wraps each NoteOn event with its NoteOff event in a MooNote
		ArrayList noteOns, noteOffs;
		for (int i = 0; i < tracks.length; i++) {
			// Searches the sequence for NoteOn and NoteOff events
			noteOns = new ArrayList(tracks[i].size() / 2);
			noteOffs = new ArrayList(tracks[i].size() / 2);
			for (int j = 0; j < tracks[i].size(); j++) {
				event = tracks[i].get(j);
				if (event.getMessage().getStatus() >= 144 &&
				    event.getMessage().getStatus() < 160) noteOns.add(event);
				if (event.getMessage().getStatus() >= 128 &&
				    event.getMessage().getStatus() < 144) noteOffs.add(event);
			}
			noteOns.trimToSize();
			noteOffs.trimToSize();
			if (noteOns.size() == 0) emptyTracks.add(tracks[i]);
			
			// Sorts the note lists by tick position.
			Comparator c = new NoteComparator();
			Collections.sort(noteOns, c);
			Collections.sort(noteOffs, c);

			// Replaces each NoteOn event it with a MooNote containing a reference to the NoteOff event.
			Iterator iOn = noteOns.iterator(), iOff;
			MidiEvent on, off = null, nextOff;
			ShortMessage onMsg, nextOffMsg;
			while(iOn.hasNext()) {
				on = (MidiEvent)iOn.next();
				onMsg = (ShortMessage)on.getMessage();
				iOff = noteOffs.iterator();
				while(iOff.hasNext()) {
					nextOff = (MidiEvent)iOff.next();
					nextOffMsg = (ShortMessage)nextOff.getMessage();
					if(onMsg.getChannel() == nextOffMsg.getChannel() &&
					   onMsg.getData1() == nextOffMsg.getData1() &&
					   c.compare(nextOff, on) > 0) {
						off = nextOff;
						iOff.remove();
						break;
					}
					   	
				}
				tracks[i].remove(on);
				if (off != null) {
					tracks[i].add(new MooNote(on, off));
				} else {
					tracks[i].add(new MooNote(on, new MidiEvent((ShortMessage)on.getMessage().clone(), on.getTick() + 48)));
				}
				iOn.remove();
			}
		}
		// Sends sequence to GUI and sequencer, then returns
		if (gui != null) gui.setSequence(seq);
		try {
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		return true;
	}

	/** 
	 * Prompts the user .
	 */
	public static boolean promptOnUnsavedChanges() {
		if (!isEdited) return false;
		int exitOption = JOptionPane.showConfirmDialog(gui,
			"The current sequence has been edited, but not saved.\nDo you wish to continue anyway?",
			"File not saved - continue?", 
			JOptionPane.OK_CANCEL_OPTION, 
			JOptionPane.WARNING_MESSAGE);
		if (exitOption == JOptionPane.CANCEL_OPTION || exitOption == JOptionPane.CLOSED_OPTION) return true;
		return false;
	}

	/** 
	 * Saves the current sequence to the previously given filename.
	 */
	public static boolean save() {
		if (filename == null) return false;
		else {
			saveAs(filename);
			return true;
		}
	}

	/** 
	 * Saves the current sequence to the given filename
	 * @param file	the filename to use
	 */
	public static boolean saveAs(String file) {
		try {
			MidiSystem.write(seq, 1, new File(file));
			filename = file;
			isEdited = false;
			gui.setStatus("Saved " + file);
			return true;
		} catch (IOException e) {
			gui.setStatus("Failed in saving " + file);
			return false;
		}
	}

	/** 
	 * Releases all reserved devices and exits the program.
	 */
	public static void quit() {
		if (gui != null) {
			if (promptOnUnsavedChanges()) return;
		}
		if (sequencer.isOpen()) sequencer.close();
		if (synthesizer.isOpen()) synthesizer.close();
		System.exit(0);
	}
	
	/** 
	 * A Comparator for sorting lists of MidiEvents.
	 */
	public static class NoteComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			return (int)(((MidiEvent)o1).getTick() - ((MidiEvent)o2).getTick());
		}
	}
}
